##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tagging Engine Implementation

$Id: engine.py 104494 2009-09-24 16:45:26Z trollfot $
"""
__docformat__ = "reStructuredText"

import persistent
import zope.interface
from zope import component
from BTrees import IOBTree, OOBTree
import random

from zope.app.container import contained
from zope.app.intid.interfaces import IIntIdRemovedEvent, IIntIds
from lovely.tag import interfaces, tag
from zope.dottedname.resolve import resolve
import types

class TaggingEngine(persistent.Persistent, contained.Contained):
    zope.interface.implements(interfaces.ITaggingEngine,
                              interfaces.ITaggingStatistics)

    _v_nextid = None
    
    def __init__(self):
        super(TaggingEngine, self).__init__()
        self._reset()

    def _generateId(self):
        """Generate an id which is not yet taken.

        This tries to allocate sequential ids so they fall into the
        same BTree bucket, and randomizes if it stumbles upon a
        used one.
        """
        while True:
            if self._v_nextid is None:
                self._v_nextid = random.randrange(0, 2**31)
            uid = self._v_nextid
            self._v_nextid += 1
            if uid not in self._tagid_to_obj:
                return uid
            self._v_nextid = None

    def _add(self, tagObj):
        uid = self._generateId()
        self._tagid_to_obj[uid] = tagObj
        # set the __parent__ in order to get a _p_oid for the object
        #tagObj.__parent__ = self
        return uid

    def _reset(self):
        # mapping of tagid to tag object
        self._tagid_to_obj = IOBTree.IOBTree()
        # Our indices for efficient querying
        self._user_to_tagids = OOBTree.OOBTree()
        self._item_to_tagids = IOBTree.IOBTree()
        self._name_to_tagids = OOBTree.OOBTree()


    @property
    def tagCount(self):
        return len(self._name_to_tagids)

    @property
    def itemCount(self):
        return len(self._item_to_tagids)

    @property
    def userCount(self):
        return len(self._user_to_tagids)

    def update(self, item, user, tags):
        """See interfaces.ITaggingEngine"""
        tags_item = set(self._item_to_tagids.get(item, ()))
        tags_user = set(self._user_to_tagids.get(user, ()))
        tags_tags = set()
        for t in tags:
            tags_tags.update(self._name_to_tagids.get(t, ()))
        old_tag_ids = tags_item.intersection(tags_user)
        # any tags of the same user/item that are  in tags
        common_tag_ids = old_tag_ids.intersection(tags_tags)
            
        common_tags = set([self._tagid_to_obj[id].brain()
                        for id in old_tag_ids])

        new_tags = set([tag.Tag(item, user, tagName).brain()
                        for tagName in tags])

        add_tags = new_tags.difference(common_tags)
        
        add_tag_ids = []
        for tagBrain in add_tags:
            tagObj = tag.Tag(*tagBrain)
            id = self._add(tagObj)
            add_tag_ids.append(id)
            ids = self._user_to_tagids.get(user)
            if ids is None:
                self._user_to_tagids[user] = IOBTree.IOSet((id,))
            else:
                ids.insert(id)

            ids = self._item_to_tagids.get(item)
            if ids is None:
                self._item_to_tagids[item] = IOBTree.IOSet((id,))
            else:
                ids.insert(id)

            ids = self._name_to_tagids.get(tagObj.name)
            if ids is None:
                self._name_to_tagids[tagObj.name] = IOBTree.IOSet((id,))
            else:
                ids.insert(id)
        del_tag_ids = old_tag_ids.difference(tags_tags)
        self._delTags(del_tag_ids)

    def _delTags(self, del_tag_ids):
        """deletes tags in iterable"""
        for id in del_tag_ids:
            tagObj = self._tagid_to_obj[id]
            self._user_to_tagids[tagObj.user].remove(id)
            if not len(self._user_to_tagids[tagObj.user]):
                del self._user_to_tagids[tagObj.user]

            self._item_to_tagids[tagObj.item].remove(id)
            if not len(self._item_to_tagids[tagObj.item]):
                del self._item_to_tagids[tagObj.item]

            self._name_to_tagids[tagObj.name].remove(id)
            if not len(self._name_to_tagids[tagObj.name]):
                del self._name_to_tagids[tagObj.name]
            del self._tagid_to_obj[id]

    def delete(self, item=None, user=None, tag=None):
        tags = None
        if item is not None:
            tags = set(self._item_to_tagids.get(item, ()))
        if user is not None:
            user_tags = set(self._user_to_tagids.get(user, ()))
            if tags is not None:
                tags = tags.intersection(user_tags)
            else:
                tags = user_tags
        if tag is not None:
            name_tags = set(self._name_to_tagids.get(tag, ()))
            if tags is not None:
                tags = tags.intersection(name_tags)
            else:
                tags = name_tags
        self._delTags(tags)

    def getTags(self, items=None, users=None):
        """See interfaces.ITaggingEngine"""
        if items is None and users is None:
            # shortcut
            return set(self._name_to_tagids.keys())

        result = self.getTagObjects(items, users)
        return set([tag.name for tag in result])

    def _getTagIds(self, items=None, users=None, tags=None):
        if items is None and users is None and tags is None:
            # get them all
            result = set()
            for v in self._item_to_tagids.values():
                result.update(v)
            return result
        result = None
        for seq, bt in ((items, self._item_to_tagids),
                        (users, self._user_to_tagids),
                        (tags, self._name_to_tagids)):
            res = set()
            if seq is not None:
                for item in seq:
                    res.update(bt.get(item, set()))
                if result is not None:
                    result = result.intersection(res)
                else:
                    result = res
        return result

    def getTagObjects(self, items=None, users=None,  tags=None):
        ids = self._getTagIds(items, users, tags)
        return set([self._tagid_to_obj[id] for id in ids])

    def getItems(self, tags=None, users=None):
        """See interfaces.ITaggingEngine"""
        uids = self._getTagIds(items=None, users=users, tags=tags)
        res = set()
        for uid in uids:
            o = self._tagid_to_obj.get(uid)
            if o is not None:
                res.add(o.item)
        return res

    def getUsers(self, tags=None, items=None):
        """See interfaces.ITaggingEngine"""
        ids = self._getTagIds(items=items, users=None, tags=tags)
        return set([self._tagid_to_obj[id].user for id in ids])

    def getRelatedTags(self, tag, degree=1):
        """See interfaces.ITaggingEngine"""
        result = set()
        degree_counter = 1
        previous_degree_tags = set([tag])
        degree_tags = set()
        while degree_counter <= degree:
            for cur_name in previous_degree_tags:
                tagids = self._name_to_tagids.get(cur_name, ())
                for tagid in tagids:
                    tag_obj = self._tagid_to_obj[tagid]
                    degree_tags.update(self.getTags(
                        items=(tag_obj.item,), users=(tag_obj.user,) ))
            # After all the related tags of this degree were found, update the
            # result set and clean up the variables for the next round.
            result.update(degree_tags)
            previous_degree_tags = degree_tags
            degree_tags = set()
            degree_counter += 1
        # Make sure the original is not included
        if tag in result:
            result.remove(tag)
        return result

    def getRelatedItems(self, item):
        tags = self.getTags([item])
        items = self.getItems(tags)
        if item in items:
            items.remove(item)
        result = []
        for otherItem in items:
            otherTags = self.getTags([otherItem])
            result.append((otherItem, len(tags.intersection(otherTags))))
        return sorted(result, key=lambda i: i[1], reverse=True)

    def getRelatedUsers(self, user):
        tags = self.getTags(users=[user])
        users = self.getUsers(tags)
        if user in users:
            users.remove(user)
        result = []
        for otherUser in users:
            otherTags = self.getTags(users=[otherUser])
            result.append((otherUser, len(tags.intersection(otherTags))))
        return sorted(result, key=lambda i: i[1], reverse=True)

    def getCloud(self, items=None, users=None):
        """See interfaces.ITaggingEngine"""
        if type(items) == types.IntType:
            items = [items]
        if type(users) in types.StringTypes:
            users = [users]

        tags = self.getTagObjects(items=items, users=users)
        d = {}
        for tag in tags:
            if d.has_key(tag.name):
                d[tag.name] += 1
            else:
                d[tag.name] = 1
        return set(d.items())

    def getFrequency(self, tags):
        """See interfaces.ITaggingEngine"""
        result = {}
        for tag in tags:
            frequency = 0
            if tag in self._name_to_tagids:
                frequency = len(self._name_to_tagids[tag])
            result[tag] = frequency
        return set(result.items())

    def __repr__(self):
        return '<%s entries=%i>' %(self.__class__.__name__,
                                   len(self._tagid_to_obj))

    def cleanStaleItems(self):
        """clean out stale items which have no associated object"""
        intIds = zope.component.getUtility(IIntIds, context=self)
        cleaned = []
        for uid in  self.getItems():
            obj = intIds.queryObject(uid)
            if obj is None:
                self.delete(item=uid)
                cleaned.append(uid)
        return cleaned

    def rename(self, old, new):
        """rename tag @old to @new"""

        if old == new:
            return 0
        tagIds = set(self._name_to_tagids.get(old, ()))
        for tagId in tagIds:
            tagObj = self._tagid_to_obj[tagId]
            tagObj.name = new
        newTagIds = IOBTree.IOSet(self._name_to_tagids.get(new, ()))
        newTagIds.update(tagIds)
        self._name_to_tagids[new] = newTagIds
        del self._name_to_tagids[old]
        return len(tagIds)

    def normalize(self, normalizer):
        if type(normalizer) == types.StringType:
            normalizer = resolve(normalizer)
        names = list(self._name_to_tagids.keys())
        count = 0
        for name in names:
            newName = normalizer(name)
            if newName != name:
                count += self.rename(name, newName)
        return count


@component.adapter(IIntIdRemovedEvent)
def removeItemSubscriber(event):
    """A subscriber to IntIdRemovedEvent which removes an item from
    the tagging engine"""
    ob = event.object
    if not interfaces.ITaggable.providedBy(ob):
        return
    for engine in zope.component.getAllUtilitiesRegisteredFor(
        interfaces.ITaggingEngine, context=ob):
        uid = zope.component.getUtility(IIntIds, context=engine).queryId(ob)
        if uid is not None:
            engine.delete(uid)
            



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tagging Views

$Id: tag.py 71822 2007-01-08 17:49:28Z dobe $
"""
__docformat__ = "reStructuredText"

from zope.publisher.browser import BrowserView
from zope import component, schema
from lovely.tag.interfaces import ITaggingEngine, ITagging
from zope.cachedescriptors.property import Lazy
from zope.app.pagetemplate import ViewPageTemplateFile
from zope.formlib import form
from lovely.tag import _
import pytz
import datetime
from zope.interface.common import idatetime

class TaggingMixin(object):

    @Lazy
    def engine(self):
        return component.getUtility(ITaggingEngine)

    @Lazy
    def tagging(self):
        return ITagging(self.context)

    def getCloud(self, maxTags=100, calc=None):
        """returns a tag cloud"""
        cloud = self.engine.getCloud()
        return normalize(cloud, maxTags, 6, calc)


class TaggingView(BrowserView, TaggingMixin):
    """Show the tags of the context as a cloud"""

    cloud       = ViewPageTemplateFile('tagcloud.pt')
    linkedcloud = ViewPageTemplateFile('linkedtagcloud.pt')

    def __init__(self, context, request):
        super(TaggingView, self).__init__(context, request)


class RelatedView(BrowserView, TaggingMixin):
    """Show related tags as a cloud"""

    cloud       = ViewPageTemplateFile('tagcloud.pt')
    linkedcloud = ViewPageTemplateFile('linkedtagcloud.pt')

    def __init__(self, context, request):
        super(RelatedView, self).__init__(context, request)

    def getCloud(self, tag=None, maxTags=100, calc=None):
        """returns a tag cloud"""
        if tag is None:
            tag = self.request.get('tagname',None)
        if tag is None:
            return []
        cloud = self.engine.getFrequency(self.engine.getRelatedTags(tag))
        return normalize(cloud, maxTags, 6, calc)


def normalize(cloud, maxTags=100, maxValue=6, calc=None):
    if calc is None:
        calc = lambda x: x
    if len(cloud) == 0:
        return []
    minmax = sorted(cloud, key=lambda i: i[1],reverse=True)

    if maxTags>0:
        end = min(maxTags,len(minmax))
    else:
        end = len(minmax)
    if end == 0:
        return []
    minmax = minmax[:end]
    minFreq = calc(minmax[-1][1])
    maxFreq = calc(minmax[0][1])
    freqRange = maxFreq-minFreq
    if freqRange>0:
        ratio = float(maxValue-1)/freqRange
    else:
        ratio = None
    res = []
    for tag, frequency in sorted(minmax):
        if ratio is None:
            normalized=1
        else:
            normalized = int((calc(frequency)-minFreq)*ratio) +1
        res.append(dict(name=tag,
                        normalized=normalized,
                        frequency=frequency,))
    return res


class UserTagForm(form.EditForm, TaggingMixin):

    form_fields = form.Fields(
        schema.TextLine(__name__='tags', title=_("Tags")))

    def setUpWidgets(self, ignore_request=False):
        # override to add existing tags to the field
        tags = u' '.join(self.tagging.getTags(
            [self.request.principal.id]))
        self.widgets = form.setUpWidgets(
            self.form_fields, self.prefix, self.context, self.request,
            data=dict(tags=tags),
            ignore_request=ignore_request
            )

    @form.action(_("Apply"))
    def handle_edit_action(self, action, data):
        tags = data.get('tags','')
        tags = set(tags.split())
        user = self.request.principal.id
        oldTags = self.tagging.getTags(user)
        if oldTags != tags:
            self.tagging.update(user, tags)
            formatter = self.request.locale.dates.getFormatter(
                            'dateTime', 'medium')
            try:
                time_zone = idatetime.ITZInfo(self.request)
            except TypeError:
                time_zone = pytz.UTC
            status = _("Updated on ${date_time}",
                       mapping={'date_time':
                                formatter.format(
                                   time_zone.normalize(datetime.datetime.now())
                                )
                               }
                      )
            self.status = status
        else:
            self.status = _('No changes')





##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: interfaces.py 70651 2006-10-15 14:19:37Z jukart $
"""
__docformat__ = "reStructuredText"

from zope import interface
from zope import schema

from zope.app.container import constraints
from zope.app.container.interfaces import IContainer


class IUrl(interface.Interface):
    constraints.containers('.IUrlContainer')

    url = schema.TextLine(
            title = u'URL',
            required = True,
            )


class IUrlContainer(IContainer):
    constraints.contains(IUrl)


class IPerformanceTestSite(interface.Interface):
    """A site for the perfomance test"""




from zope.app.zopeappgenerations import getRootFolder
from zope.app.generations.utility import findObjectsProviding
from lovely.tag.interfaces import ITaggingEngine
from BTrees import IOBTree
from zope.app.component.interfaces import ISite

def evolve(context):

    """evolve to use an IOBTree instead of PersistentList"""
    for s in findObjectsProviding(getRootFolder(context), ISite):
        for engine in s.getSiteManager().getAllUtilitiesRegisteredFor(
            ITaggingEngine):

            if hasattr(engine, '_tags'):
                engine._tagid_to_obj = IOBTree.IOBTree()
                for uid, ref in engine._tag_ids.items():
                    obj = ref()
                    engine._tagid_to_obj[uid] = obj
                del engine._tags
                del engine._tag_ids
                



import unittest
from zope.app.testing import functional

try:
    from z3c import sampledata
    sampledata = True
except ImportError:
    sampledata = False

functional.defineLayer('TestLayer', 'ftesting.zcml')


def setUp(test):
    """Setup a reasonable environment for the tag tests"""
    pass


def tearDown(test):
    pass


def test_suite():
    suite = unittest.TestSuite()
    if not sampledata:
        return suite
    suites = (
        functional.FunctionalDocFileSuite('browser/README.txt',
                                          setUp=setUp, tearDown=tearDown,
                                          ),
        )
    for s in suites:
        s.layer=TestLayer
        suite.addTest(s)
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tag Implementation

$Id: tagging.py 71642 2006-12-21 17:27:38Z dobe $
"""
__docformat__ = "reStructuredText"
import zope.component
import zope.interface
from zope.app import intid
from zope.cachedescriptors.property import Lazy
from zope.security.management import getInteraction
from lovely.tag import interfaces


class Tagging(object):
    zope.interface.implements(interfaces.ITagging)
    zope.component.adapts(interfaces.ITaggable)

    engineName = ''

    def __init__(self, context):
        self.context = context

    @Lazy
    def docId(self):
        ids = zope.component.getUtility(intid.interfaces.IIntIds,
                                        context=self.engine)
        id = ids.queryId(self.context)
        if id is None:
            ids.register(self.context)
            id = ids.getId(self.context)
        return id

    @Lazy
    def engine(self):
        return zope.component.getUtility(interfaces.ITaggingEngine,
                                         context=self.context,
                                         name=self.engineName)

    def update(self, user, tags):
        """See interfaces.ITagging"""
        try:
            self.engine.update(self.docId, user, tags)
        except zope.component.ComponentLookupError:
            # special behaviour :
            #  If we update without tags it is possible that we do this
            #  because an object has been deleted. This is usually done in an
            #  event handler for ObjectRemovedEvent. If we would raise an
            #  exeption in this case it is not possible to delete a site.
            if tags:
                raise

    def getTags(self, users=None):
        """See interfaces.ITagging"""
        return self.engine.getTags(items=(self.docId,), users=users)

    def getUsers(self, tags=None):
        """See interfaces.ITagging"""
        return self.engine.getUsers(items=(self.docId,), tags=tags)


class UserTagging(object):
    zope.interface.implements(interfaces.IUserTagging)
    zope.component.adapts(interfaces.ITaggable)

    engineName = ''

    def __init__(self, context):
        self.context = context

    @Lazy
    def docId(self):
        ids = zope.component.getUtility(intid.interfaces.IIntIds)
        id = ids.queryId(self.context)
        if id is None:
            ids.register(self.context)
            id = ids.getId(self.context)
        return id

    @Lazy
    def engine(self):
        return zope.component.getUtility(interfaces.ITaggingEngine,
                                         name=self.engineName)

    @property
    def _pid(self):
        participations = getInteraction().participations
        if participations:
            return participations[0].principal.id
        else:
            raise ValueError, "User not found"

    @apply
    def tags():
        def fget(self):
            return self.engine.getTags(items=(self.docId,),
                                        users=(self._pid,))
        def fset(self, value):
            if value is None:
                return
            try:
                self.engine.update(self.docId, self._pid, value)
            except zope.component.ComponentLookupError:
                # special behaviour :
                #  If we update without tags it is possible that we do this
                #  because an object has been deleted. This is usually done in an
                #  event handler for ObjectRemovedEvent. If we would raise an
                #  exeption in this case it is not possible to delete a site.
                if value:
                    raise
        return property(**locals())




##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This is not Free or Open Source software, but its source is made available
# when it is purchased. Please see the author(s) for details about its 
# license and what you are permitted to do with its source code.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""The performance site configurator.

$Id: configurator.py 70651 2006-10-15 14:19:37Z jukart $
"""
__docformat__ = "reStructuredText"

from zope import component
from zope import event
from zope.lifecycleevent import ObjectCreatedEvent

from zope.app.component import hooks
from zope.app.component.interfaces import ISite
from zope.app.component.site import LocalSiteManager
from zope.app.intid import IntIds
from zope.app.intid.interfaces import IIntIds

from z3c.configurator import configurator

from lovely import tag

from lovely.tag.performance.interfaces import IPerformanceTestSite


class SetUpPerformanceTestSite(configurator.ConfigurationPluginBase):
    component.adapts(IPerformanceTestSite)

    def __call__(self, data):
        if not ISite.providedBy(self.context):
            sm = LocalSiteManager(self.context)
            self.context.setSiteManager(sm)
        else:
            sm = self.context.getSiteManager()
        hooks.setSite(self.context)

        default = sm['default']

        if 'intid' not in default:
            intid = IntIds()
            event.notify(ObjectCreatedEvent(intid))
            default['intid'] = intid
            sm.registerUtility(intid, IIntIds)

        # Add the tagging engine
        if 'tagging-engine' not in default:
            engine = tag.TaggingEngine()
            event.notify(ObjectCreatedEvent(engine))
            default['tagging-engine'] = engine
            sm.registerUtility(engine, tag.interfaces.ITaggingEngine)




##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tag Implementation

$Id: tag.py 104494 2009-09-24 16:45:26Z trollfot $
"""
__docformat__ = "reStructuredText"

import datetime
import persistent
import zope.interface
from zope.schema.fieldproperty import FieldProperty

from lovely.tag import interfaces

def normalizeTagName(name):
    return name.lower()

class Tag(persistent.Persistent):
    """Simple implementation of a tag."""
    zope.interface.implements(interfaces.ITag)

    item = FieldProperty(interfaces.ITag['item'])
    user = FieldProperty(interfaces.ITag['user'])
    name = FieldProperty(interfaces.ITag['name'])
    timestamp = FieldProperty(interfaces.ITag['timestamp'])
    
    def __init__(self, item, user, name):
        self.item = item
        if not isinstance(user, unicode):
            user = unicode(user, 'utf-8')
        self.user = user
        self.name = name
        self.timestamp = datetime.datetime.now()

    def __cmp__(self, other):
        return cmp((self.item, self.user, self.name),
                   (other.item, other.user, other.name))

    def brain(self):
        """ representation to build sets"""
        return (self.item, self.user, self.name)

    def __repr__(self):
        return '<%s %r for %i by %r>' %(
            self.__class__.__name__, self.name, self.item, self.user)



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Sample Data for Tagging

$Id: sampledata.py 70516 2006-10-04 08:21:50Z jukart $
"""
import os
import random

from zope import interface
from zope import component
from zope import schema
from zope.app.component import hooks

from z3c.sampledata.interfaces import ISampleDataPlugin
from lovely import tag

LOCAL_DIR = os.path.dirname(__file__)

from lovely import tag


def generate(max_tags=100000, engine=None, seed=0):

    rand = random.Random(seed)

    tag_data = file(os.path.join(LOCAL_DIR, 'tags.data'), 'r').read().split()
    tag_data = [unicode(name) for name in tag_data]

    user_data = file(os.path.join(LOCAL_DIR, 'users.data'), 'r').read().split()
    user_data = [unicode(user) for user in user_data]

    if engine is None:
        engine = tag.TaggingEngine()

    cur_tags = 0
    obj_num = 0
    while cur_tags < max_tags:
        obj_num += 1
        user_num = rand.randint(20, 30)
        for user in rand.sample(user_data, user_num):
            tags_num = rand.randint(1, 7)
            cur_tags += tags_num
            tags = rand.sample(tag_data, tags_num)
            engine.update(obj_num, user, tags)

    return engine


class SampleEngine(object):
    interface.implements(ISampleDataPlugin)
    name = 'lovely.tags.sampleengine'
    dependencies = []
    schema = None

    def generate(self, context, param={}, dataSource=None, seed=None):
        hooks.setSite(context)

        engine = component.queryUtility(tag.interfaces.ITaggingEngine)
        if engine is None:
            # Add the tagging engine
            engine = tag.TaggingEngine()
            sm = context.getSiteManager()
            sm['default']['tagging-engine'] = engine
            component.provideUtility(engine, tag.interfaces.ITaggingEngine)
        return engine


class ISampleTags(interface.Interface):
    """Parameters for the sample tag generator"""

    numTags = schema.Int(
            title = u'Number Of Tags',
            required = True,
            default = 20,
            )


class SampleTags(object):
    interface.implements(ISampleDataPlugin)
    name = 'lovely.tags.sampledata'
    dependencies = []
    schema = ISampleTags

    def generate(self, context, param={}, dataSource=None, seed=None):
        engine = component.getUtility(tag.interfaces.ITaggingEngine)
        return generate(param['numTags'], engine, seed)




##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This is not Free or Open Source software, but its source is made available
# when it is purchased. Please see the author(s) for details about its
# license and what you are permitted to do with its source code.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""The event handlers for the performance tests.

$Id: event.py 70651 2006-10-15 14:19:37Z jukart $
"""
__docformat__ = "reStructuredText"

from zope import component

from zope.app.container.interfaces import IObjectAddedEvent

from z3c.configurator import configurator

from lovely.tag.performance.interfaces import IPerformanceTestSite


@component.adapter(IPerformanceTestSite, IObjectAddedEvent)
def onPerformanceTestSiteAdded(site,event):
    configurator.configure(site, {})




##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tag test setup

$Id: tests.py 73440 2007-03-22 13:15:53Z dobe $
"""
__docformat__ = "reStructuredText"

import doctest
import unittest
from zope.app.testing import setup
from zope.testing.doctestunit import DocFileSuite

def setUp(test):
    setup.placefulSetUp()

def tearDown(test):
    setup.placefulTearDown()

def test_suite():

    stressSuite = DocFileSuite(
        'stresstest.txt',
        setUp=setUp, tearDown=tearDown,
        optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS)
    stressSuite.level = 2

    suite = unittest.TestSuite(
        (
        DocFileSuite('README.txt',
                     setUp=setUp,
                     tearDown=tearDown,
                     optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
                     ),
        DocFileSuite('index.txt',),

        ))

    suite.addTest(stressSuite)
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Sampledata for the performance test

$Id: sampledata.py 73827 2007-03-28 08:53:55Z dobe $
"""
__docformat__ = 'restructuredtext'

import os
import random

from zope import interface
from zope import component
from zope import schema
from zope import event

from zope.lifecycleevent import ObjectCreatedEvent

from z3c.sampledata.interfaces import ISampleDataPlugin

from lovely.tag.interfaces import IUserTagging
from lovely.tag.performance import app


class IUrlsSchema(interface.Interface):
    """Sample generator for urls with tags."""

    numUrls = schema.Int(
            title = u'urls',
            description = u'Number or urls to create.',
            default = 100,
            )


class Urls(object):
    interface.implements(ISampleDataPlugin)

    dependencies = []
    schema = IUrlsSchema

    def generate(self, context, param, dataSource=None, seed=None):
        numUrls = param['numUrls']
        urls = app.UrlContainer()
        event.notify(ObjectCreatedEvent(urls))
        context['urls'] = urls
        dirname = os.path.dirname(__file__)
        self.tags = [unicode(line.strip())
                         for line in file(os.path.join(dirname,
                                          '40000.words')).readlines()]
        self.shorttags = [unicode(line.strip())
                         for line in file(os.path.join(dirname,
                                          '250.words')).readlines()]
        self.domains = [unicode(line.strip())
                         for line in file(os.path.join(dirname,
                                          'domain.txt')).readlines()]
        self.protocol = [unicode(line.strip())
                         for line in file(os.path.join(dirname,
                                          'protocol.txt')).readlines()]
        self.pre = [unicode(line.strip())
                         for line in file(os.path.join(dirname,
                                          'pre.txt')).readlines()]
        self.rand = random.Random()
        self.rand.seed(seed)
        i = 1
        for urlName in self._urls(numUrls):
            url = app.Url()
            event.notify(ObjectCreatedEvent(url))
            url.url = urlName
            urls['%i'%i] = url
            tagging = IUserTagging(url)
            tagging.tags=self._tags(10)
            i+=1
        return urls

    def _tags(self, num, randomLength=True):
        tags = []
        maxRand = len(self.tags)-1
        maxShortRand = len(self.shorttags)-1
        if randomLength:
            num = self.rand.randint(0, num)
        for i in range(num):
            if divmod(i,2)[1]==0:
                tags.append(self.tags[self.rand.randint(0, maxRand)])
            else:
                tags.append(self.shorttags[self.rand.randint(0, maxShortRand)])
        return tags

    def _urls(self, num):
        maxTagsRand = len(self.tags)-1
        for i in range(num):
            protocol = self.protocol[self.rand.randint(0, len(self.protocol)-1)]
            pre      = self.pre[self.rand.randint(0, len(self.pre)-1)]
            firstPart = self.tags[self.rand.randint(0, maxTagsRand)]
            secondPart = self.tags[self.rand.randint(0, maxTagsRand)]
            domain = self.domains[self.rand.randint(0, len(self.domains)-1)]
            thirdPart = self.tags[self.rand.randint(0, maxTagsRand)]
            forthPart = self.tags[self.rand.randint(0, maxTagsRand)]
            yield u'%s%s.%s.%s/%s/%s.html' %(
                        protocol,
                        pre,
                        firstPart,
                        domain,
                        thirdPart,
                        forthPart,
                        )


class IPerformanceTestSiteSamples(interface.Interface):
    """Sample generator for the users bookshelfs."""

    name = schema.TextLine(
            title = u'Name',
            description = u'The name of the site.',
            default = u'Tag Performance'
            )


class PerformanceTestSite(object):
    interface.implements(ISampleDataPlugin)

    dependencies = []
    schema = IPerformanceTestSiteSamples

    def generate(self, context, param, dataSource=None, seed=None):
        name = param['name']
        testsite = app.PerformanceTestSite()
        event.notify(ObjectCreatedEvent(testsite))
        context[name] = testsite
        return testsite



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A catalog index wich uses the tagging engine.

$Id: index.py 78118 2007-07-18 16:58:38Z schwendinger $
"""
__docformat__ = "reStructuredText"

import persistent
from BTrees.IFBTree import IFTreeSet

from zope import interface
from zope import component

from zope.cachedescriptors.property import Lazy
from zope.index.interfaces import IInjection

from zope.app.container.contained import Contained

from lovely.tag.interfaces import ITaggingEngine, ITagIndex


class TagIndex(persistent.Persistent, Contained):
    interface.implements(ITagIndex, IInjection)

    def __init__(self, engineName=''):
        self.engineName = engineName

    @Lazy
    def engine(self):
        return component.getUtility(ITaggingEngine, name = self.engineName)

    def index_doc(self, docid, value):
        # not used but makes the catalog happy
        pass

    def unindex_doc(self, docid):
        # not used but makes the catalog happy
        pass

    def apply(self, query):
        if 'any_of' in query:
            tags = query['any_of']
            items = self.engine.getItems(tags)
        elif 'all_of' in query:
            tags = query['all_of']
            items = None
            for tag in tags:
                it = self.engine.getItems([tag])
                if items is None:
                    items = it
                else:
                    items = items.intersection(it)
                if not items:
                    break
        if items:
            items = [id for id in items]
            return IFTreeSet(items)
        return IFTreeSet()

    def clear(self):
        pass



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: app.py 70651 2006-10-15 14:19:37Z jukart $
"""
__docformat__ = "reStructuredText"

import persistent
from zope import interface

from zope.schema.fieldproperty import FieldProperty

from zope.app.container.contained import Contained
from zope.app.container import btree
from zope.app.folder.folder import Folder

from lovely.tag.interfaces import ITaggable
from lovely.tag.performance.interfaces import IUrl, IUrlContainer, IPerformanceTestSite


class PerformanceTestSite(Folder):
    interface.implements(IPerformanceTestSite)

    def __repr__(self):
        return '<%s %r>' %(self.__class__.__name__, self.__name__)


class Url(persistent.Persistent, Contained):
    interface.implements(IUrl, ITaggable)

    url = FieldProperty(IUrl['url'])


class UrlContainer(btree.BTreeContainer):
    interface.implements(IUrlContainer)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)




from zope.app.zopeappgenerations import getRootFolder
from zope.app.generations.utility import findObjectsProviding
from lovely.tag.interfaces import ITaggingEngine
from BTrees import IOBTree
from zope.app.component.interfaces import ISite
def evolve(context):
    """evolve to use -IOSet- instead of -set- instances"""

    for s in findObjectsProviding(getRootFolder(context), ISite):
        for engine in s.getSiteManager().getAllUtilitiesRegisteredFor(
            ITaggingEngine):
            for key, val in engine._name_to_tagids.items():
                changed = False
                for uid in list(val):
                    if engine._tag_ids.queryObject(uid) is None:
                        changed=True
                        val.remove(uid)
                if changed or not isinstance(val, IOBTree.IOSet):
                    engine._name_to_tagids[key] = IOBTree.IOSet(val)

                



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Interfaces

$Id: interfaces.py 72966 2007-03-03 14:47:07Z jukart $
"""
__docformat__ = "reStructuredText"

import zope.interface
import zope.schema

class ITag(zope.interface.Interface):
    """A single tag."""

    item = zope.schema.Int(
        title=u'Item',
        description=u'The item that is tagged.',
        required=True)

    user = zope.schema.TextLine(
        title=u'User',
        description=u'The user id that created the tag.',
        required=True)

    name = zope.schema.TextLine(
        title=u'Tag Name',
        description=u'The tag name the user provided for the item.',
        required=True)

    timestamp = zope.schema.Datetime(
        title=u'Timestamp',
        description=u"The timestamp of the tag's creation date",
        required=True)


class ITaggingEngine(zope.interface.Interface):
    """Manangement and Querying of Tags.

    Tags are small stickers that are specific to an object to be tageed and
    the user tagging the object.
    """

    def update(item, user, tags):
        """Update the tagging engine for an item and user and its tags.

        The item must be an integer, the user can be a string and the tags
        argument must be an iterable of tag names (strings).

        This method always overwrites the old tag settings. However, existing
        tags will not be readded and are just skipped.
        """

    def getTags(items=None, users=None):
        """Get all tags matching the specified items and users.

        If no items are specified, match all items. Ditto for users.

        The method returns a set of *normalized* tag names.
        """

    def getTagObjects(self, items, users):
        """same as getTags but returns tag objects implementing
        ITag"""

    def getCloud(items=None, users=None):
        """Get a set of tuples in the form of ('tag',
        frequency). Arguments are the same as getTags."""

    def getItems(tags=None, users=None):
        """Get all items matching the specified items and users.

        If no tags are specified, match all tags. Ditto for users.

        The method returns a set of item ids.
        """

    def getUsers(tags=None, items=None):
        """Get all items matching the specified tags and items.

        If no tags are specified, match all tags. Ditto for items.

        The method returns a set of user strings.
        """

    def getRelatedTags(tag, degree=1):
        """Get a set of all related tags."""

    def getRelatedItems(item):
        """Get a list of items related to item

        Items are related if they have a least one tag in common with `item`.
        The result contains a list of tuples in the form (item, numTags).
        numTags is the number of tags in common. The result is reverse sorted
        by the number of tags in common.
        """

    def getRelatedUsers(user):
        """Get a list of users related a user

        Users are related if they have a least one tag in common with `user`.
        The result contains a list of tuples in the form (user, numTags).
        numTags is the number of tags in common. The result is reverse sorted
        by the number of tags in common.
        """

    def getFrequency(tags):
        """Get the frequency of all tags

        Returns tuples in the form of ('tag', frequency)
        """

    def cleanStaleItems(self):
        """clean out stale items which have no associated object.

        A list of the cleaned out uids is returned"""

    def delete(item=None, user=None, tag=None):

        """Globall delete all tag entries filtered by the given
        criteria, if a criteria is None it is ignored

        e.g. to delete all entries of a user do:
        delete(user=u'username')
        """

    def rename(old, new):
        """rename tags from @old to @new, this method joins the tags
        if tags with the new name do exist"""

    def normalize(normalizer):
        """Normalize tagnames with the given normalizer
        function. @normalizer can also be a dotted name.

        The function provided should return a new name for an existing
        name."""


class ITaggingStatistics(zope.interface.Interface):
    """A tagging engine that provides statistical information about itself"""

    tagCount = zope.schema.Int(
            title = u'Tags',
            description = u'The number of tags in the tagging engine',
            required = True,
            )

    itemCount = zope.schema.Int(
            title = u'Items',
            description = u'The number of items in the tagging engine',
            required = True,
            )

    userCount = zope.schema.Int(
            title = u'Users',
            description = u'The number of users in the tagging engine',
            required = True,
            )


class ITaggable(zope.interface.Interface):
    """A marker interface to declare an object as being able to be tagged."""


class ITagging(zope.interface.Interface):
    """Provides a tagging API for a single item.

    Effectively, this interface provides a simplification (special case) of
    the tagging engine API.
    """

    def update(user, tags):
        """Update the tags of an item.

        See ``ITaggingEngine.update()`` for more information.
        """

    def getTags(users=None):
        """Get all tags matching the specified users.

        See ``ITaggingEngine.getTags()`` for more information.
        """

    def getUsers(tags=None):
        """Get all items matching the specified tags.

        See ``ITaggingEngine.getUsers()`` for more information.
        """

class IUserTagging(zope.interface.Interface):

    """Provides easy tagging of objects based on the current
    principal"""

    tags = zope.schema.Set(title=u'Tags',
                           description=u'Tags for the current User',
                           required=False)


class ITagIndex(zope.interface.Interface):

    def apply(query):
        """Return None or an IFTreeSet of the doc ids that match the query.

        query is a dict with one of the following keys: and, or

        Any one of the keys may be used; using more than one is not allowed.

        'any_of' : docs containing at least one of the keys are returned. The
                   value is a list containing the tags.
        'all_of' : docs containing at alll of the tags are returned. The value
                   is a list containing the tags.
        """



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tagging Views

$Id: engine.py 73827 2007-03-28 08:53:55Z dobe $
"""
__docformat__ = "reStructuredText"

from zope.publisher.browser import BrowserView
from zope import schema
from zope.formlib import form
from lovely.tag import _
from zope.formlib import form
import csv
import cStringIO

class ManageView(form.PageForm):

    label=_(u"Manage Tagging Engine")
    
    form_fields = form.Fields(
        schema.DottedName(__name__=u'normalizer',
                          title=_(u'Normalizer Function'),
                          required=False),
        schema.TextLine(__name__=u'oldName',
                        title=_(u'Old Name'),
                        required=False),
        schema.TextLine(__name__=u'newName',
                        title=_(u'New Name'),
                        required=False),
        )
    
    @form.action(label=_(u'Clean Stale Items'))
    def cleanStale(self, action, data):
        cleaned = self.context.cleanStaleItems()
        self.status = u'Cleaned out %s items' % len(cleaned)
        

    @form.action(label=_(u'Normalize'))
    def normalize(self, action, data):
        normalizer = data.get('normalizer')
        if not normalizer:
            self.status=_(u'No normalizer function defined')
            return
        count = self.context.normalize(normalizer)
        self.status = u'Normalized %s tag objects' % count

    @form.action(label=_(u'Rename Tag'))
    def renameTag(self, action, data):
        oldName = data.get('oldName')
        newName = data.get('newName')
        if not (oldName and newName):
            self.status=_(u'Please define old and new name.')
            return
        count = self.context.rename(oldName, newName)
        self.status = u'Renamed %s tag objects' % count


class CSVExportView(BrowserView):

    def __call__(self):
        self.request.response.setHeader('Content-Type', 'text/csv')
        res = cStringIO.StringIO()
        writer = csv.writer(res, dialect=csv.excel)
        encoding = 'utf-8'
        writer.writerow(('Name', 'User', 'Item', 'Timestamp'))
        for tag in self.context.getTagObjects():
            row = [tag.name.encode(encoding)]
            row.append(tag.user.encode(encoding))
            row.append(tag.item)
            row.append(tag.timestamp)
            writer.writerow(row)
        return res.getvalue()

        
        



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tagging Performance Views

$Id: views.py 70651 2006-10-15 14:19:37Z jukart $
"""
__docformat__ = "reStructuredText"

import time
import random
import os

from zope import interface
from zope import component
from zope import schema

from zope.formlib import form

from zope.app.pagetemplate import ViewPageTemplateFile
from zope.app.intid.interfaces import IIntIds

from lovely.tag.interfaces import ITaggingEngine


class ISearchSchema(interface.Interface):
    """Defines the search fields"""

    tags = schema.TextLine(
            title = u'Tags',
            description = u"""
                The tags to search for.
                If no tags are given :
                    Search for random tags from the file '250.words'.
                    First search with 1 tag, second with 2 tags, third with 3
                    tags and then start again with one tag.
                """,
            required = False,
            )

    repeats = schema.Int(
            title = u'Repeat',
            description = u'Number of times to repeat the search',
            default = 300,
            )

    itemLookup = schema.Bool(
            title = u'Item lookup',
            description = u"""
                Read the items after geting the ids from the tagging engine
                """,
            default = False,
            )


class SearchTags(form.PageForm):

    form_fields = form.Fields(ISearchSchema).select('tags',
                                                    'repeats',
                                                    'itemLookup')

    base_template = form.AddForm.template
    template = ViewPageTemplateFile('searchtags.pt')

    hasTimeValues = False

    @form.action(u'Search')
    def handle_search_action(self, action, data):
        self.rand = random.Random()
        self.rand.seed('tags')
        dirname = os.path.dirname(__file__)
        self.shorttags = [unicode(line.strip())
                         for line in file(os.path.join(dirname,
                                          '250.words')).readlines()]
        start = time.clock()
        self.repeats = data['repeats']
        self.tags = data['tags']
        if self.tags is not None:
            self.tags = data['tags'].split()
        itemLookup = data['itemLookup']
        intids = component.getUtility(IIntIds)
        engine = component.getUtility(ITaggingEngine)
        numTags = 1
        for i in xrange(self.repeats):
            if self.tags is None:
                tags = self.tags
            else:
                if numTags>3:
                    numTags=1
                tags = self._tags(numTags)
                numTags+=1
            itemids = engine.getItems(tags)
            self.numitemsfound = len(itemids)
            self.itemsfound = []
            if itemLookup:
                for itemId in itemids:
                    self.itemsfound.append(intids.getObject(itemId))
        self.totaltime = time.clock() - start
        self.time = self.totaltime / self.repeats
        self.hasTimeValues = True

    def _tags(self, num):
        tags = []
        maxShortRand = len(self.shorttags)-1
        for i in range(num):
            tags.append(self.shorttags[self.rand.randint(0, maxShortRand)])
        return tags

